package org.processmining.lpengines.implementations;

import java.util.HashMap;
import java.util.Map;

import org.processmining.framework.util.Pair;
import org.processmining.lpengines.interfaces.LPEngine;
import org.processmining.lpengines.interfaces.LPEngineVisitor;

import lpsolve.LpSolve;
import lpsolve.LpSolveException;

public class LPEngineLpSolveImpl implements LPEngine {

	static private boolean loaded = false;

	static {
		if (!loaded) {
			init();
		}
	}

	public static void init() {
		System.loadLibrary("lpsolve55");
		System.loadLibrary("lpsolve55j");
		loaded = true;
	}

	// define underlying LP-solver engineType and load libraries (static)
	protected LpSolve engine;

	protected final Object lock = new Object();

	public LPEngineLpSolveImpl() {
		try {
			this.engine = LpSolve.makeLp(0, 0);
			setup();
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
	}

	public LPEngineLpSolveImpl(int rows, int columns) {
		try {
			this.engine = LpSolve.makeLp(rows, columns);
			setup();
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
	}

	public void accept(LPEngineVisitor visitor) {
		visitor.visit(this);
	}

	@Override
	public int addConstraint(double[] constraint, Operator operator, double value) {
		assert (constraint.length == numVariables() + 1);
		int rowIndex = numConstraints() + 1;
		try {
			this.engine.addConstraint(constraint, this.mapLPSolveOperator(operator), value);
		} catch (LpSolveException e) {
			e.printStackTrace();
			rowIndex = -1;
		}
		return rowIndex;
	}

	@Override
	public int addConstraint(Map<Integer, Double> constraint, Operator operator, double value) {
		return addConstraint(variableMapToArray(constraint), operator, value);
	}

	@Override
	public int addVariable(double[] column, VariableType type) {
		assert (variableMatchesType(column, type));
		assert (column.length == this.engine.getNrows() + 1);
		int newVariable = numVariables() + 1;
		try {
			this.engine.addColumn(column);
			setType(newVariable, type);
		} catch (LpSolveException e) {
			e.printStackTrace();
			return -1;
		}
		return newVariable;
	}

	@Override
	public int addVariable(Map<Integer, Double> nonZeroValues, VariableType type) {
		assert (variableMatchesType(nonZeroValues, type));
		int newVariable = numVariables() + 1;
		Pair<int[], double[]> nonZeroArrayPair = mapToIndexValueArrayPair(nonZeroValues);
		try {
			this.engine.addColumnex(nonZeroValues.size(), nonZeroArrayPair.getSecond(), nonZeroArrayPair.getFirst());
			setType(newVariable, type);
		} catch (LpSolveException e) {
			e.printStackTrace();
			return -1;
		}
		return newVariable;
	}

	protected boolean boolCheck(double value) {
		return (value >= 0d && value <= 1d && value == Math.floor(value) && !Double.isInfinite(value));
	}

	@Override
	public LPEngine clone() throws CloneNotSupportedException {
		LPEngineLpSolveImpl result;
		try {
			result = new LPEngineLpSolveImpl(1, 1);
			LpSolve clone;
			synchronized (lock) {
				clone = engine.copyLp();
			}
			result.setEngine(clone);
		} catch (LpSolveException e) {
			System.out.println(e.toString());
			result = null;
		}
		return result;
	}

	public void destroy() {
		engine.deleteAndRemoveLp();
	}

	@Override
	public double[] emptyConstraint() {
		return new double[numVariables() + 1];
	}

	@Override
	public double[] emptyVariable() {
		return new double[numConstraints() + 1];
	}

	@Override
	public EngineType engineType() {
		return EngineType.LPSOLVE;
	}

	public double getVariableLowerBound(int variableIndex) {
		double result = 0.0d;
		try {
			result = this.engine.getLowbo(variableIndex);
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
		return result;
	}

	public double getVariableUpperBound(int variableIndex) {
		double result = 0.0d;
		try {
			result = this.engine.getUpbo(variableIndex);
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
		return result;
	}

	protected boolean intCheck(double value) {
		return (value == Math.floor(value) && !Double.isInfinite(value));
	}

	@Override
	public boolean isFeasible() {
		boolean result = false;
		try {
			LpSolve clone;
			synchronized (lock) {
				clone = engine.copyLp();
			}
			clone.setBreakAtFirst(true);
			if (clone.solve() == 0) {
				result = true;
				clone.deleteAndRemoveLp();
			}
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
		return result;
	}

	/**
	 * Translates the LPEngine Operators to LpSolve operators.
	 * 
	 * @param operator
	 *            defined in LPEngine
	 * @return corresponding LpSolve Operator
	 */
	private int mapLPSolveOperator(Operator operator) {
		int result;
		switch (operator) {
			case LESS_EQUAL :
				result = LpSolve.LE;
				break;
			case EQUAL :
				result = LpSolve.EQ;
				break;
			case GREATER_EQUAL :
				result = LpSolve.GE;
				break;
			default :
				return LpSolve.EQ;
		}
		return result;
	}

	protected Pair<int[], double[]> mapToIndexValueArrayPair(Map<Integer, Double> map) {
		int[] indexes = new int[map.size()];
		double[] values = new double[map.size()];

		int arrayIndex = 0;
		for (int index : map.keySet()) {
			indexes[arrayIndex] = index;
			values[arrayIndex] = map.get(index);
			arrayIndex++;
		}
		return new Pair<int[], double[]>(indexes, values);
	}

	@Override
	public int minimalConstraintIndex() {
		return 1;
	}

	@Override
	public int minimalVariableIndex() {
		return 1;
	}

	@Override
	public int numConstraints() {
		return this.engine.getNorigRows();
	}

	@Override
	public int numVariables() {
		return this.engine.getNorigColumns();
	}

	@Override
	public void pop() {
		try {
			this.engine.delConstraint(this.numConstraints());
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
	}

	@Override
	public void print() {
		this.engine.printLp();
	}

	@Override
	public boolean removeConstraint(int row) {
		assert (row <= numConstraints());
		boolean result = false;
		try {
			this.engine.delConstraint(row);
			result = true;
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
		return result;
	}

	@Override
	public boolean removeVariable(int column) {
		assert (column <= numVariables());
		boolean result = false;
		try {
			this.engine.delColumn(column);
			result = true;
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
		return result;
	}

	@Override
	public boolean setConstraint(int row, double[] constraint) {
		assert (row <= this.engine.getNrows());
		boolean result = false;
		try {
			this.engine.setRow(row, constraint);
			result = true;
		} catch (LpSolveException e) {
			e.printStackTrace();
			result = false;
		}
		return result;
	}

	public boolean setConstraint(int row, double[] constraint, double rhs) {
		boolean result = false;
		try {
			engine.setRow(row, constraint);
			engine.setRh(row, rhs);
			result = true;
		} catch (LpSolveException e) {
			e.printStackTrace();
			result = false;
		}
		return result;
	}

	public boolean setConstraint(int row, double[] constraint, Operator op, double rhs) {
		boolean result = false;
		try {
			engine.setRow(row, constraint);
			engine.setConstrType(row, mapLPSolveOperator(op));
			engine.setRh(row, rhs);
			result = true;
		} catch (LpSolveException e) {
			e.printStackTrace();
			result = false;
		}
		return result;
	}

	@Override
	public boolean setConstraint(int row, Map<Integer, Double> constraint) {
		return setConstraint(row, variableMapToArray(constraint));
	}

	@Override
	public boolean setConstraintOperator(int row, Operator op) {
		assert (row <= this.engine.getNrows());
		boolean result = false;
		try {
			this.engine.setConstrType(row, this.mapLPSolveOperator(op));
			result = true;
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
		return result;
	}

	public void setEngine(LpSolve engine) {
		this.engine = engine;
	}

	@Override
	public boolean setLowerBound(int column, double lowerBound) {
		assert (column <= numVariables());
		boolean result = true;
		try {
			this.engine.setLowbo(column, lowerBound);
		} catch (LpSolveException e) {
			e.printStackTrace();
			result = false;
		}
		return result;
	}

	@Override
	public void setMat(int rowIndex, int columnIndex, double value) {
		assert (rowIndex <= this.engine.getNrows());
		assert (columnIndex <= this.engine.getNcolumns());
		try {
			this.engine.setMat(rowIndex, columnIndex, value);
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
	}

	@Override
	public boolean setObjective(double[] objective, ObjectiveTargetType target) {
		assert (objective.length == numVariables() + 1);
		try {
			this.engine.setObjFn(objective);
			setObjectiveType(target);
		} catch (LpSolveException e) {
			e.printStackTrace();

		}
		return true;
	}

	@Override
	public boolean setObjective(Map<Integer, Double> objective, ObjectiveTargetType target) {
		try {
			this.engine.setObjFn(this.variableMapToArray(objective));
			setObjectiveType(target);
			return true;
		} catch (LpSolveException e) {
			e.printStackTrace();
			return false;
		}
	}

	@Override
	public boolean setObjectiveType(ObjectiveTargetType target) {
		switch (target) {
			case MAX :
				this.engine.setMaxim();
				break;
			case MIN :
				this.engine.setMinim();
				break;
		}
		return true;
	}

	public boolean setRightHandSide(int row, double rhs) {
		boolean success = false;
		try {
			engine.setRh(row, rhs);
			success = true;
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
		return success;
	}

	public void setTimeOut(long seconds) {
		engine.setTimeout(seconds);
	}

	@Override
	public boolean setType(int column, VariableType type) {
		try {
			// set the column fresh
			this.engine.setUnbounded(column);
			this.engine.setLowbo(column, 0);
			this.engine.setInt(column, false);
			switch (type) {
				case INTEGER :
					this.engine.setInt(column, true);
					break;
				case BOOLEAN :
					this.engine.setInt(column, true);
					this.engine.setUpbo(column, 1);
					this.engine.setLowbo(column, 0);
					break;
				case REAL :
					break;
				default :
					break;
			}

		} catch (LpSolveException e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}

	public void setup() {
		if (numVariables() > 0 || numConstraints() > 0) {
			this.engine.setAddRowmode(true);
		}
		this.engine.setScaling(LpSolve.SCALE_GEOMETRIC | LpSolve.SCALE_EQUILIBRATE);// | LpSolve.SCALE_INTEGERS);
		this.engine.setScalelimit(5);
		this.engine.setPivoting(LpSolve.PRICER_DEVEX | LpSolve.PRICE_ADAPTIVE);
		this.engine.setMaxpivot(250);
		this.engine.setBbFloorfirst(LpSolve.BRANCH_AUTOMATIC);
		this.engine.setBbRule(LpSolve.NODE_PSEUDONONINTSELECT | LpSolve.NODE_GREEDYMODE | LpSolve.NODE_DYNAMICMODE
				| LpSolve.NODE_RCOSTFIXING);
		this.engine.setBbDepthlimit(-50);
		this.engine.setAntiDegen(LpSolve.ANTIDEGEN_FIXEDVARS | LpSolve.ANTIDEGEN_STALLING);
		this.engine.setImprove(LpSolve.IMPROVE_DUALFEAS | LpSolve.IMPROVE_THETAGAP);
		this.engine.setBasiscrash(LpSolve.CRASH_NOTHING);
		this.engine.setSimplextype(LpSolve.SIMPLEX_DUAL_PRIMAL);
		this.engine.setVerbose(1);
	}

	@Override
	public boolean setUpperBound(int variable, double upperBound) {
		assert (variable <= numVariables());
		try {
			this.engine.setUpbo(variable, upperBound);
			return true;
		} catch (LpSolveException e) {
			e.printStackTrace();
			return false;
		}
	}

	@Override
	public Map<Integer, Double> solve() {
		Map<Integer, Double> result = new HashMap<Integer, Double>();
		LpSolve clone = null;
		try {
			synchronized (lock) {
				clone = this.engine.copyLp();
			}
			if (clone.solve() == 0) {
				// get a pointer to the result and store it.
				double[] solution = clone.getPtrVariables();
				for (int variable = 0; variable < solution.length; variable++) {
					if (solution[variable] > 0) {
						result.put(variable + 1, solution[variable]);
					}
				}
			}
		} catch (LpSolveException e) {
			e.printStackTrace();
			return result;
		} finally {
			if (clone != null) {
				clone.deleteAndRemoveLp();
			}
		}
		return result;
	}

	public Pair<double[], Double> solveAndResetAndValueNative() {
		double[] solution = new double[0];
		Double objectiveValue = null;
		try {
			synchronized (lock) {
				if (engine.solve() == 0) {
					double[] raw = engine.getPtrVariables();
					solution = new double[raw.length + 1];
					System.arraycopy(raw, 0, solution, 1, raw.length);
					objectiveValue = engine.getObjective();
				}
			}
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
		engine.defaultBasis();
		return new Pair<double[], Double>(solution, objectiveValue);
	}

	@Override
	public Pair<Map<Integer, Double>, Double> solveAndValue() {
		Map<Integer, Double> solution = new HashMap<Integer, Double>();
		double value;
		Pair<Map<Integer, Double>, Double> result = new Pair<Map<Integer, Double>, Double>(solution, null);
		LpSolve clone = null;
		try {
			synchronized (lock) {
				clone = this.engine.copyLp();
			}
			if (clone.solve() == 0) {
				double[] rawSol = clone.getPtrVariables();
				for (int variable = 0; variable < rawSol.length; variable++) {
					if (rawSol[variable] > 0) {
						solution.put(variable + 1, rawSol[variable]);
					}
				}
				value = clone.getObjective();
				result = new Pair<Map<Integer, Double>, Double>(solution, value);
			}
		} catch (LpSolveException e) {
			e.printStackTrace();
			result = new Pair<Map<Integer, Double>, Double>(new HashMap<Integer, Double>(), null);
		} finally {
			if (clone != null) {
				clone.deleteAndRemoveLp();
			}
		}
		return result;
	}

	@Override
	public Pair<double[], Double> solveAndValueNative() {
		double[] solution = new double[0];
		Double objectiveValue = null;
		LpSolve clone = null;
		try {
			synchronized (lock) {
				clone = this.engine.copyLp();
			}
			if (clone.solve() == 0) {
				double[] raw = clone.getPtrVariables();
				solution = new double[raw.length + 1];
				System.arraycopy(raw, 0, solution, 1, raw.length);
				objectiveValue = clone.getObjective();
			}
		} catch (LpSolveException e) {
			e.printStackTrace();
		} finally {
			if (clone != null) {
				clone.deleteAndRemoveLp();
			}
		}
		return new Pair<double[], Double>(solution, objectiveValue);
	}

	@Override
	public double[] solveNative() {
		double[] solution = new double[0];
		LpSolve clone = null;
		try {
			synchronized (lock) {
				clone = this.engine.copyLp();
			}
			if (clone.solve() == 0) {
				double[] raw = clone.getPtrVariables();
				solution = new double[raw.length + 1];
				System.arraycopy(raw, 0, solution, 1, raw.length);
			}
		} catch (LpSolveException e) {
			e.printStackTrace();
		} finally {
			clone.deleteAndRemoveLp();
		}
		return solution;
	}

	public Pair<Pair<double[], Double>, Long> solveValueNativeSolveTimeNano() {
		double[] solution = new double[0];
		Double objectiveValue = null;
		LpSolve clone = null;
		long startNs;
		long solveTimeNs = -1;
		try {
			synchronized (lock) {
				clone = this.engine.copyLp();
			}
			startNs = System.nanoTime();
			int res = clone.solve();
			solveTimeNs = System.nanoTime() - startNs;
			if (res == 0) {
				double[] raw = clone.getPtrVariables();
				solution = new double[raw.length + 1];
				System.arraycopy(raw, 0, solution, 1, raw.length);
				objectiveValue = clone.getObjective();
			}
		} catch (LpSolveException e) {
			e.printStackTrace();
		} finally {
			if (clone != null) {
				clone.deleteAndRemoveLp();
			}
		}
		return new Pair<Pair<double[], Double>, Long>(new Pair<double[], Double>(solution, objectiveValue),
				solveTimeNs);
	}

	protected boolean valueMatchesType(double d, VariableType type) {
		boolean result;
		switch (type) {
			case BOOLEAN :
				result = boolCheck(d);
				break;
			case INTEGER :
				result = intCheck(d);
				break;
			default :
				result = true;
				break;
		}
		return result;
	}

	/**
	 * translates a variable map (i.e. variableIndex-value map) to a LpSolve
	 * compliant array of doubles.
	 * 
	 * @param variableMap
	 * @return
	 */
	protected double[] variableMapToArray(Map<Integer, Double> variableMap) {
		double[] variableMapArray = new double[this.engine.getNcolumns() + 1];
		for (int col = 1; col <= numVariables(); col++) {
			if (variableMap.containsKey(col)) {
				variableMapArray[col] = variableMap.get(col);
			}
		}
		return variableMapArray;
	}

	protected boolean variableMatchesType(double[] values, VariableType type) {
		boolean result = true;
		for (int i = 1; i < values.length; i++) {
			result = valueMatchesType(values[i], type);
			if (!result)
				break;
		}
		return result;
	}

	protected boolean variableMatchesType(Map<Integer, Double> nonZeroValues, VariableType type) {
		boolean result = true;
		for (double d : nonZeroValues.values()) {
			result = valueMatchesType(d, type);
			if (!result)
				break;
		}
		return result;
	}

	@Override
	public void writeToFile(String dest) {
		try {
			engine.writeLp(dest);
		} catch (LpSolveException e) {
			e.printStackTrace();
		}
	}

}
